@@ -2,7 +2,9 @@ import contextlib
 import json
 import os
 import re
-from distutils.version import Version
+import sys
+from io import StringIO
+from signal import SIGINT, signal
 
 import kubernetes
 import urllib3
@@ -15,10 +17,12 @@ from openshift.dynamic.exceptions import (
     ServerTimeoutError,
 )
 from openshift.dynamic.resource import ResourceField
+from packaging.version import Version
 
 from ocp_resources.constants import (
     NOT_FOUND_ERROR_EXCEPTION_DICT,
     PROTOCOL_ERROR_EXCEPTION_DICT,
+    TIMEOUT_4MINUTES,
 )
 from ocp_resources.logger import get_logger
 from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler
@@ -27,13 +31,15 @@ from ocp_resources.utils import TimeoutExpiredError, TimeoutSampler
 DEFAULT_CLUSTER_RETRY_EXCEPTIONS = {
     ConnectionAbortedError: [],
     ConnectionResetError: [],
-    InternalServerError: ["etcdserver: leader changed"],
+    InternalServerError: [
+        "etcdserver: leader changed",
+        "etcdserver: request timed out",
+    ],
     ServerTimeoutError: [],
 }
 
 LOGGER = get_logger(name=__name__)
-TIMEOUT = 240
-MAX_SUPPORTED_API_VERSION = "v1"
+MAX_SUPPORTED_API_VERSION = "v2"
 
 
 def _collect_instance_data(directory, resource_object):
@@ -128,6 +134,8 @@ def _get_api_version(dyn_client, api_group, kind):
         log = f"Couldn't find {kind} in {api_group} api group"
         LOGGER.warning(log)
         raise NotImplementedError(log)
+
+    LOGGER.info(f"kind: {kind} api version: {res.group_version}")
     return res.group_version
 
 
@@ -156,7 +164,7 @@ class KubeAPIVersion(Version):
     def __init__(self, vstring=None):
         self.vstring = vstring
         self.version = None
-        super().__init__(vstring=vstring)
+        super().__init__(version=vstring)
 
     def parse(self, vstring):
         components = [comp for comp in self.component_re.split(vstring) if comp]
@@ -271,14 +279,17 @@ class Resource:
         APIREGISTRATION_K8S_IO = "apiregistration.k8s.io"
         APP_KUBERNETES_IO = "app.kubernetes.io"
         APPS = "apps"
+        BATCH = "batch"
         CDI_KUBEVIRT_IO = "cdi.kubevirt.io"
         CONFIG_OPENSHIFT_IO = "config.openshift.io"
         CONSOLE_OPENSHIFT_IO = "console.openshift.io"
         EVENTS_K8S_IO = "events.k8s.io"
+        FLAVOR_KUBEVIRT_IO = "flavor.kubevirt.io"
         FORKLIFT_KONVEYOR_IO = "forklift.konveyor.io"
         HCO_KUBEVIRT_IO = "hco.kubevirt.io"
         HOSTPATHPROVISIONER_KUBEVIRT_IO = "hostpathprovisioner.kubevirt.io"
         IMAGE_OPENSHIFT_IO = "image.openshift.io"
+        IMAGE_REGISTRY = "registry.redhat.io"
         K8S_CNI_CNCF_IO = "k8s.cni.cncf.io"
         K8S_V1_CNI_CNCF_IO = "k8s.v1.cni.cncf.io"
         KUBERNETES_IO = "kubernetes.io"
@@ -294,6 +305,7 @@ class Resource:
         )
         NETWORKING_ISTIO_IO = "networking.istio.io"
         NETWORKING_K8S_IO = "networking.k8s.io"
+        NODE_LABELLER_KUBEVIRT_IO = "node-labeller.kubevirt.io"
         NMSTATE_IO = "nmstate.io"
         NODEMAINTENANCE_KUBEVIRT_IO = "nodemaintenance.kubevirt.io"
         OPERATOR_OPENSHIFT_IO = "operator.openshift.io"
@@ -302,6 +314,7 @@ class Resource:
         OS_TEMPLATE_KUBEVIRT_IO = "os.template.kubevirt.io"
         PACKAGES_OPERATORS_COREOS_COM = "packages.operators.coreos.com"
         POLICY = "policy"
+        POOL_KUBEVIRT_IO = "pool.kubevirt.io"
         PROJECT_OPENSHIFT_IO = "project.openshift.io"
         RBAC_AUTHORIZATION_K8S_IO = "rbac.authorization.k8s.io"
         RIPSAW_CLOUDBULLDOZER_IO = "ripsaw.cloudbulldozer.io"
@@ -320,6 +333,7 @@ class Resource:
         UPLOAD_CDI_KUBEVIRT_IO = "upload.cdi.kubevirt.io"
         V2V_KUBEVIRT_IO = "v2v.kubevirt.io"
         VM_KUBEVIRT_IO = "vm.kubevirt.io"
+        MIGRATIONS_KUBEVIRT_IO = "migrations.kubevirt.io"
 
     class ApiVersion:
         V1 = "v1"
@@ -332,9 +346,13 @@ class Resource:
         name=None,
         client=None,
         teardown=True,
-        timeout=TIMEOUT,
+        timeout=TIMEOUT_4MINUTES,
         privileged_client=None,
         yaml_file=None,
+        delete_timeout=TIMEOUT_4MINUTES,
+        dry_run=None,
+        node_selector=None,
+        node_selector_labels=None,
     ):
         """
         Create a API resource
@@ -352,8 +370,6 @@ class Resource:
         self.privileged_client = privileged_client
         self.yaml_file = yaml_file
         self.resource_dict = None  # Filled in case yaml_file is not None
-        self.logger = get_logger(name=f"{__name__.rsplit('.')[0]} {self.kind}")
-
         if not (self.name or self.yaml_file):
             raise ValueError("name or yaml file is required")
 
@@ -366,7 +382,7 @@ class Resource:
                 kubernetes.config.ConfigException,
                 urllib3.exceptions.MaxRetryError,
             ):
-                self.logger.error(
+                LOGGER.error(
                     "You need to be logged into a cluster or have $KUBECONFIG env configured"
                 )
                 raise
@@ -377,6 +393,18 @@ class Resource:
 
         self.teardown = teardown
         self.timeout = timeout
+        self.delete_timeout = delete_timeout
+        self.dry_run = dry_run
+        self.node_selector = node_selector
+        self.node_selector_labels = node_selector_labels
+        self.node_selector_spec = self._prepare_node_selector_spec()
+        self.res = None
+
+    def _prepare_node_selector_spec(self):
+        if self.node_selector:
+            return {f"{self.ApiGroup.KUBERNETES_IO}/hostname": self.node_selector}
+        if self.node_selector_labels:
+            return self.node_selector_labels
 
     @ClassProperty
     def kind(cls):  # noqa: N805
@@ -390,10 +418,16 @@ class Resource:
             dict: Resource dict.
         """
         if self.yaml_file:
-            with open(self.yaml_file, "r") as stream:
-                self.resource_dict = yaml.safe_load(stream=stream.read())
-                self.name = self.resource_dict["metadata"]["name"]
-                return self.resource_dict
+            if isinstance(self.yaml_file, StringIO):
+                data = self.yaml_file.read()
+            else:
+                with open(self.yaml_file, "r") as stream:
+                    data = stream.read()
+
+            self.resource_dict = yaml.safe_load(stream=data)
+            self.resource_dict.get("metadata", {}).pop("resourceVersion", None)
+            self.name = self.resource_dict["metadata"]["name"]
+            return self.resource_dict
 
         return {
             "apiVersion": self.api_version,
@@ -408,27 +442,41 @@ class Resource:
         return self._base_body()
 
     def __enter__(self):
+        signal(SIGINT, self._sigint_handler)
         return self.deploy()
 
     def __exit__(self, exception_type, exception_value, traceback):
+        # For debug, export kind (class name) + NOTEARDOWN to skip resource teardown.
+        # For example:export VirtualMachineInstanceNOTEARDOWN=True
+        skip_teardown_env = f"{self.kind}NOTEARDOWN"
+        no_teardown_from_environment = os.environ.get(skip_teardown_env)
+        if no_teardown_from_environment:
+            LOGGER.warning(
+                f"Skip teardown. Got {skip_teardown_env}={no_teardown_from_environment}"
+            )
+            return
+
         if self.teardown:
             self.clean_up()
 
-    def deploy(self):
-        self.create()
+    def _sigint_handler(self, signal_received, frame):
+        self.__exit__(exception_type=None, exception_value=None, traceback=None)
+        sys.exit(signal_received)
+
+    def deploy(self, wait=False):
+        self.create(wait=wait)
         return self
 
     def clean_up(self):
         if os.environ.get("CNV_TEST_COLLECT_LOGS", "0") == "1":
             try:
-                self.logger.info(f"Collecting data for {self.kind} {self.name}")
                 _collect_data(resource_object=self)
             except Exception as exception_:
-                self.logger.warning(
+                LOGGER.warning(
                     f"Log collector failed to collect info for {self.kind} {self.name}\nexception: {exception_}"
                 )
 
-        self.delete(wait=True, timeout=self.timeout)
+        self.delete(wait=True, timeout=self.delete_timeout)
 
     @classmethod
     def _prepare_resources(cls, dyn_client, singular_name, *args, **kwargs):
@@ -478,7 +526,7 @@ class Resource:
     def api(self):
         return self.full_api()
 
-    def wait(self, timeout=TIMEOUT, sleep=1):
+    def wait(self, timeout=TIMEOUT_4MINUTES, sleep=1):
         """
         Wait for resource
 
@@ -489,7 +537,7 @@ class Resource:
         Raises:
             TimeoutExpiredError: If resource not exists.
         """
-        self.logger.info(f"Wait until {self.kind} {self.name} is created")
+        LOGGER.info(f"Wait until {self.kind} {self.name} is created")
         samples = TimeoutSampler(
             wait_timeout=timeout,
             sleep=sleep,
@@ -503,7 +551,7 @@ class Resource:
             if sample:
                 return
 
-    def wait_deleted(self, timeout=TIMEOUT):
+    def wait_deleted(self, timeout=TIMEOUT_4MINUTES):
         """
         Wait until resource is deleted
 
@@ -513,7 +561,7 @@ class Resource:
         Raises:
             TimeoutExpiredError: If resource still exists.
         """
-        self.logger.info(f"Wait until {self.kind} {self.name} is deleted")
+        LOGGER.info(f"Wait until {self.kind} {self.name} is deleted")
         return self.client_wait_deleted(timeout=timeout)
 
     @property
@@ -543,7 +591,9 @@ class Resource:
             if not sample:
                 return
 
-    def wait_for_status(self, status, timeout=TIMEOUT, stop_status=None, sleep=1):
+    def wait_for_status(
+        self, status, timeout=TIMEOUT_4MINUTES, stop_status=None, sleep=1
+    ):
         """
         Wait for resource to be in status
 
@@ -556,7 +606,7 @@ class Resource:
             TimeoutExpiredError: If resource in not in desire status.
         """
         stop_status = stop_status if stop_status else self.Status.FAILED
-        self.logger.info(f"Wait for {self.kind} {self.name} status to be {status}")
+        LOGGER.info(f"Wait for {self.kind} {self.name} status to be {status}")
         samples = TimeoutSampler(
             wait_timeout=timeout,
             sleep=sleep,
@@ -582,16 +632,15 @@ class Resource:
 
         except TimeoutExpiredError:
             if current_status:
-                self.logger.error(
-                    f"Status of {self.kind} {self.name} is {current_status}"
-                )
+                LOGGER.error(f"Status of {self.kind} {self.name} is {current_status}")
             raise
 
-    def create(self, wait=False):
+    def create(self, body=None, wait=False):
         """
         Create resource.
 
         Args:
+            body (dict): Resource data to create.
             wait (bool) : True to wait for resource status.
 
         Returns:
@@ -601,20 +650,33 @@ class Resource:
             ValueMismatch: When body value doesn't match class value
         """
         data = self.to_dict()
-        self.logger.info(f"Create {self.kind} {self.name}")
-        self.logger.info(f"Posting {data}")
-        self.logger.debug(f"\n{yaml.dump(data)}")
-        res = self.api.create(body=data, namespace=self.namespace)
+        if body:
+            kind = body["kind"]
+            name = body.get("name")
+            api_version = body["apiVersion"]
+            if kind != self.kind:
+                raise ValueMismatch(f"{kind} != {self.kind}")
+            if name and name != self.name:
+                raise ValueMismatch(f"{name} != {self.name}")
+            if api_version != self.api_version:
+                raise ValueMismatch(f"{api_version} != {self.api_version}")
+
+            data.update(body)
+
+        LOGGER.info(f"Create {self.kind} {self.name}")
+        LOGGER.info(f"Posting {data}")
+        LOGGER.debug(f"\n{yaml.dump(data)}")
+        res = self.api.create(body=data, namespace=self.namespace, dry_run=self.dry_run)
         if wait and res:
             return self.wait()
         return res
 
-    def delete(self, wait=False, timeout=TIMEOUT, body=None):
-        self.logger.info(f"Delete {self.kind} {self.name}")
+    def delete(self, wait=False, timeout=TIMEOUT_4MINUTES, body=None):
+        LOGGER.info(f"Delete {self.kind} {self.name}")
         if self.exists:
             data = self.instance.to_dict()
-            self.logger.info(f"Deleting {data}")
-            self.logger.debug(f"\n{yaml.dump(data)}")
+            LOGGER.info(f"Deleting {data}")
+            LOGGER.debug(f"\n{yaml.dump(data)}")
 
         try:
             res = self.api.delete(name=self.name, namespace=self.namespace, body=body)
@@ -635,7 +697,7 @@ class Resource:
         Returns:
            str: Status
         """
-        self.logger.info(f"Get {self.kind} {self.name} status")
+        LOGGER.info(f"Get {self.kind} {self.name} status")
         return self.instance.status.phase
 
     def update(self, resource_dict):
@@ -645,8 +707,8 @@ class Resource:
         Args:
             resource_dict: Resource dictionary
         """
-        self.logger.info(f"Update {self.kind} {self.name}:\n{resource_dict}")
-        self.logger.debug(f"\n{yaml.dump(resource_dict)}")
+        LOGGER.info(f"Update {self.kind} {self.name}:\n{resource_dict}")
+        LOGGER.debug(f"\n{yaml.dump(resource_dict)}")
         self.api.patch(
             body=resource_dict,
             namespace=self.namespace,
@@ -658,14 +720,15 @@ class Resource:
         Replace resource metadata.
         Use this to remove existing field. (update() will only update existing fields)
         """
-        self.logger.info(f"Replace {self.kind} {self.name}: \n{resource_dict}")
-        self.logger.debug(f"\n{yaml.dump(resource_dict)}")
+        LOGGER.info(f"Replace {self.kind} {self.name}: \n{resource_dict}")
+        LOGGER.debug(f"\n{yaml.dump(resource_dict)}")
         self.api.replace(body=resource_dict, name=self.name, namespace=self.namespace)
 
     @staticmethod
     def retry_cluster_exceptions(
         func, exceptions_dict=DEFAULT_CLUSTER_RETRY_EXCEPTIONS, **kwargs
     ):
+
         sampler = TimeoutSampler(
             wait_timeout=10,
             sleep=1,
@@ -738,7 +801,7 @@ class Resource:
         Raises:
             TimeoutExpiredError: If Pod condition in not in desire status.
         """
-        self.logger.info(
+        LOGGER.info(
             f"Wait for {self.kind}/{self.name}'s '{condition}' condition to be '{status}'"
         )
         samples = TimeoutSampler(
@@ -807,9 +870,11 @@ class NamespacedResource(Resource):
         namespace=None,
         client=None,
         teardown=True,
-        timeout=TIMEOUT,
+        timeout=TIMEOUT_4MINUTES,
         privileged_client=None,
         yaml_file=None,
+        delete_timeout=TIMEOUT_4MINUTES,
+        **kwargs,
     ):
         super().__init__(
             name=name,
@@ -818,6 +883,8 @@ class NamespacedResource(Resource):
             timeout=timeout,
             privileged_client=privileged_client,
             yaml_file=yaml_file,
+            delete_timeout=delete_timeout,
+            **kwargs,
         )
         self.namespace = namespace
         if not (self.name and self.namespace) and not self.yaml_file:
